/*
 This file is part of QxBF.

 QxBF, a preprocessor for QBF based on failed literal detection.  
 Copyright 2011 Florian Lonsing, Aina Niemetz and Mathias Preiner.

 See also http://fmv.jku.at/qxbf for further information.

 QxBF is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or (at
 your option) any later version.

 QxBF is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with QxBF.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef INCLUDE_QXBF_H
#define INCLUDE_QXBF_H

#include <limits.h>

#define QXBF_ABORT(cond, msg)                             \
  do                                                      \
  {                                                       \
    if (cond)                                             \
    {                                                     \
      fprintf (stderr, "[QxBF] %s: %s\n", __func__, msg); \
      fflush (stderr);                                    \
      cleanup ();                                         \
      if (in != NULL)                                     \
        fclose (in);                                      \
      abort ();                                           \
    }                                                     \
  } while (0)


/* parser */
#define QDIMACS_COMMENT 'c'
#define QDIMACS_PREAMBLE 'p'
#define QDIMACS_C 'c'
#define QDIMACS_N 'n'
#define QDIMACS_F 'f'
#define QDIMACS_FORALL 'a'
#define QDIMACS_EXISTS 'e'
#define QDIMACS_EOL '0'
#define MINUS  '-'
#define EOL 0
#define PARSER_STACKSIZE 16000


#define PARSER_SKIP_SPACE_DO_WHILE(c)  \
  do                                   \
  {                                    \
    c = getc (in);                     \
  }                                    \
  while (isspace (c));

#define PARSER_SKIP_SPACE_WHILE(c) \
  while (isspace (c)) c = getc (in);

#define PARSER_READ_NUM(num, c, msg)   \
  do                                   \
  {                                    \
    num = 0;                           \
    for (;;)                           \
    {                                  \
      if (!isdigit (c))                \
        QXBF_ABORT (1, msg);           \
      num = num * 10 + (c - '0');      \
      if (isspace (c = getc (in)))     \
        break;                         \
    }                                  \
  } while (0)

/* stack */
#define STACK_INIT(stack, stack_size)                           \
  do                                                            \
  {                                                             \
    stack.bp = (int *) malloc (stack_size * sizeof (int));      \
    QXBF_ABORT (stack.bp == NULL, "memory allocation failed");  \
    memset (stack.bp, 0, stack_size * sizeof (int));            \
    stack.sp = stack.bp;                                        \
    stack.size = stack_size;                                    \
  } while (0)

#define STACK_CNT(stack) ((stack).sp - (stack).bp)

#define STACK_RESET(stack) (stack.sp = stack.bp)

#define PUSH(stack, num)                                              \
  do                                                                  \
    {                                                                 \
      *(stack.sp) = num;                                              \
      stack.sp += 1;                                                  \
      if (stack.sp - stack.bp >= stack.size)                          \
  {                                                                   \
    stack.size <<= 1;                                                 \
    stack.bp = (int *) realloc (stack.bp, stack.size * sizeof (int)); \
    QXBF_ABORT (stack.bp == NULL, "memory reallocation failed");      \
    stack.sp = stack.bp + (stack.size >> 1 );                         \
  }                                                                   \
    } while (0)


#define PUSH_L0(stack, num, l1_ptr) \
  do                                \
    {                               \
      PUSH (stack, num);            \
      l1_ptr += 1;                  \
    }                               \
  while (0)


#define POP(stack, num)  \
  do                     \
    {                    \
      stack.sp -= 1;     \
      num = *(stack.sp); \
      *(stack.sp) = 0;   \
    }                    \
  while(0)

typedef int Lit;
typedef int VarId;
typedef int ClauseId;

#define INVALID_CLAUSE_ID INT_MAX
#define INVALID_TRAIL_POS UINT_MAX

typedef enum
{
  QTYPE_EXISTS = -1,
  QTYPE_UNDEF = 0,
  QTYPE_FORALL = 1
} QType;


typedef enum
{
  STATE_UNDEF = -1,
  STATE_SAT = 0,
  STATE_UNSAT = 1,
  STATE_CONFLICT = 2
} QxBFState;


typedef enum
{
  BTYPE_UNDEF = -1,
  BTYPE_FALSE = 0,
  BTYPE_TRUE = 1
} BType;


typedef enum
{
  DLEVEL_UNDEF = -1,
  DLEVEL_0 = 0,
  DLEVEL_1 = 1
} DLevel;


typedef enum
{
  VMODE_UNDEF,
  VMODE_UNIT,
  VMODE_PURE
} VMode;

typedef struct
{
  int size;
  int *bp;                      /* base pointer */
  int *sp;                      /* stack pointer */
} Stack;


typedef struct
{
  VarId id;
  QType type;
  VMode mode;
  DLevel d_level;               /* decision level  */
  int s_level;                  /* scope level  */
  BType assignment;
  unsigned int is_propagated:1;
  int num_pos_occs;
  int num_pos_occs_cur;
  int num_neg_occs;
  int num_neg_occs_cur;
  ClauseId *pos_occs;
  ClauseId *neg_occs;
  /* For unit literals: store clause which caused 'var' to be assigned
     as unit. */
  ClauseId antecedent;
  /* Assignment order of variables, i.e. position on stack
     'assigned_vars' which is called 'trail'. */
  unsigned int trail_pos;
  unsigned int sat_test_pos_ignore:1;
  unsigned int sat_test_neg_ignore:1;
  unsigned int learnt_from_unsat_test:1;
} Var;


typedef struct
{
  ClauseId id;
  int num_lits;
  Lit *lits;
  VarId sat_var;                /* id of first var to satisfy clause */
  unsigned int is_pure_sat:1;
  unsigned int is_tautology:1;
} Clause;


typedef struct
{
  QType type;
  int level;
  int num_vars;
  VarId *vars;
} Scope;


typedef struct
{
  unsigned int no_pures:1;
  unsigned int seed;
  unsigned int verbosity;
  unsigned int bcp_prop_limit;
  unsigned int init_bcp_prop_limit;
  unsigned int sat_dec_limit;
  unsigned int init_sat_dec_limit;
  unsigned int sat_ignore_dist;
  unsigned int sat_sweep_calls;
  unsigned int sat_sweep_dec_limit;
  unsigned int sat_time_limit;
  unsigned int bcp_time_limit;
  unsigned int sat_dec_limit_inc;
  double sat_time_threshold;
  /* If 'no_abs' is true, then never treat outer universal variables
     as existential ones. Instead use q-resolution to check if detected
     conflicts were valid. */
  unsigned int no_abs:1;
  unsigned int no_bcp:1;
  unsigned int no_sat:1;
  unsigned int simple_output:1;
  unsigned int full_clean_output:1;
  unsigned int print_toplevel:1;
} QxBFOptions;


typedef struct
{
  unsigned int propagations;
  unsigned int decisions;
  unsigned int units;
  unsigned int pure_literals;
} QxBFStatistics;


inline static void assign (VarId, BType, DLevel);
static void update_occs_cnt (ClauseId, int);
static int is_clause_empty (ClauseId);
static int is_clause_unit (ClauseId);
static void cleanup (void);


#ifndef NDEBUG
static void dump (void);
static void assert_check_pure_literals (int);
static void assert_check_occurences (void);
static void assert_check_var_scope_levels (void);
static void assert_check_is_propagated (void);
#endif

#endif
